
import logging
import json

try:
    import cPickle as p
except ImportError:
    import pickle as p

LOGGER = logging.getLogger(__name__)


class Indexer(object):
    def __init__(self, pad='~', oov='±', verbose=False):
        """
        Parameters:
        -----------
        reserved: dict {int: any}, preindexed reserved items.
        This is useful in case of reserved values (e.g. padding).
        Indices must start at 0.

        Example:
        --------
        indexer = Indexer()
        """
        self.level = logging.WARN if verbose else logging.NOTSET
        self.pad = pad
        self.oov = oov
        self.decoder = {}
        self.encoder = {}
        self._current = 0
        if pad:
            self.pad = pad
            self.pad_code = self.encode(pad, fitted=False)
        if oov:
            self.oov = oov
            self.oov_code = self.encode(oov, fitted=False)

    def vocab(self):
        return self.encoder.keys()

    def vocab_len(self):
        return len(self.encoder)

    def set_verbose(self, verbose=True):
        self.level = logging.WARN if verbose else logging.NOTSET

    def fit(self, seq):
        self.fitted = True
        return self.encode_seq(seq, fitted=False)

    def transform(self, seq):
        if not self.fitted:
            raise ValueError("Indexer hasn't been fitted yet")
        return self.encode_seq(seq, fitted=True)

    def fit_transform(self, seq):
        return self.fit(seq)

    def encode(self, s, fitted=True):
        """
        Parameters:
        -----------
        s: object, object to index

        Returns:
        --------
        idx (int)
        """
        if s in self.encoder:
            return self.encoder[s]
        elif fitted:
            if self.oov:
                return self.oov_code
            else:
                raise KeyError("Unknown value [%s] with no OOV default" % s)
        else:
            LOGGER.log(self.level, "Inserting new item [%s]" % s)
            idx = self._current
            self.encoder[s] = idx
            self.decoder[idx] = s
            self._current += 1
            return idx

    def decode(self, idx):
        return self.decoder[idx]

    def encode_seq(self, seq, **kwargs):
        return [self.encode(x, **kwargs) for x in seq]

    def _to_json(self):
        obj = {'encoder': self.encoder, 'decoder': self.decoder}
        if self.pad:
            obj.update({'pad': self.pad, 'pad_code': self.pad_code})
        if self.oov:
            obj.update({'oov': self.oov, 'oov_code': self.oov_code})
        return obj

    def save(self, fname, mode='json'):
        if mode == 'json':
            with open(fname, 'w') as f:
                json.dump(self._to_json(), f)
        elif mode == 'pickle':
            with open(fname, 'wb') as f:
                p.dump(self, f)
        else:
            raise ValueError('Unrecognized mode %s' % mode)

    @staticmethod
    def load(fname, mode='json'):
        if mode == 'pickle':
            with open(fname, 'rb') as f:
                return p.load(f)
        elif mode == 'json':
            with open(fname, 'r') as f:
                return Indexer.from_dict(json.load(f))
        else:
            raise ValueError('Unrecognized mode %s' % mode)

    @classmethod
    def from_dict(cls, d, **kwargs):
        idxr = cls(**kwargs)
        idxr.encoder = d['encoder']
        idxr.decoder = {int(k): v for (k, v) in d['decoder'].items()}
        idxr.fitted = True
        idxr._current = len(idxr.encoder)
        if 'pad' in d:
            idxr.pad = d['pad']
            idxr.pad_code = idxr.encode(idxr.pad, fitted=True)
        if 'oov' in d:
            idxr.oov = d['oov']
            idxr.oov_code = idxr.encode(idxr.oov, fitted=True)
        return idxr




import numpy as np
import inspect
from time import time

from scipy import random
from sklearn import svm
from sklearn.cross_validation import train_test_split
from sklearn.metrics import accuracy_score

from casket.experiment import Experiment


def generate_clusters(n_points=100, n_clusters=5, n_dim=3, noise=0.15):
    X = np.zeros((n_points * n_clusters, n_dim))
    y = np.zeros(n_points * n_clusters)
    for cluster in range(n_clusters):
        mean = (np.ones(n_dim) * (cluster + 1)) + (random.rand(n_dim))
        random_mat = random.rand(n_dim, n_dim)
        cov = np.dot(random_mat, random_mat.transpose())  # pos semi-definite
        sample = np.random.multivariate_normal(mean, cov, n_points)
        sample += (noise * np.random.random_sample((n_points, n_dim)))
        X[cluster * n_points: (cluster + 1) * n_points] = sample
        y[cluster * n_points: (cluster + 1) * n_points] = cluster
    return X, y


class Clustering(Experiment):
    def get_id(self):
        return inspect.getsourcefile(lambda: 0)

if __name__ == '__main__':
    X, y = generate_clusters(noise=0.5, n_clusters=4)
    X_train, X_test, y_train, y_test = \
        train_test_split(X, y, test_size=0.33, random_state=42)
    clf = svm.LinearSVC()

    model_meta = {}
    model = Clustering.use("test.db", tags=('random', 'test')) \
                      .model("LinearSVC", model_meta=model_meta)
    with model.session(clf.get_params(), ensure_unique=False) as session:
        start = time()
        clf.fit(X_train, y_train)
        end = time()
        session.add_meta({"duration": end - start})

        y_pred = clf.predict(X_test)
        session.add_result({"y_pred": y_pred.tolist(),
                            "y_test": y_test.tolist(),
                            "accuracy": accuracy_score(y_test, y_pred)})




import logging
from subprocess import check_output, CalledProcessError

from . import utils


logger = logging.getLogger(__name__)


def log(msg, level=logging.WARN):
    logger.log(level, msg)


class GitInfo:
    """
    Utility class to retrieve git-based info from a repository
    """
    def __init__(self, fname):
        self.dirname = utils.get_dir(fname)

    def run(self, cmd):
        try:
            output = check_output(cmd, cwd=self.dirname)
            return output.strip().decode('utf-8')
        except OSError:
            log("Git doesn't seem to be installed in your system")
        except CalledProcessError:
            log("Not a git repository")

    def get_commit(self):
        """
        Returns current commit on file or None if an error is thrown by git
        (OSError) or if file is not under git VCS (CalledProcessError)
        """
        return self.run(["git", "describe", "--always"])

    def get_branch(self):
        """
        Returns current active branch on file or None if an error is thrown
        by git (OSError) or if file is not under git VCS (CalledProcessError)
        """
        return self.run(["git", "rev-parse", "--abbrev-ref", "HEAD"])

    def get_tag(self):
        """
        Returns current active tag
        """
        return self.run(["git", "describe", "--tags", "--abbrev=0"])




import inspect
import os
from datetime import datetime
from contextlib import contextmanager
import sys


@contextmanager
def silence(stderr=True, stdout=True):
    with open(os.devnull, "w") as devnull:
        old_stderr = sys.stderr
        old_stdout = sys.stdout
        sys.stderr = devnull
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stderr = old_stderr
            sys.stdout = old_stdout


def get_dir(fname):
    if os.path.isfile(fname):
        return os.path.dirname(fname)
    elif os.path.isdir(fname):
        return fname
    else:
        raise ValueError("Not valid path %s" % fname)


def getsourcefile(f):
    """
    Returns the file in which function `f` was defined.
    """
    return os.path.realpath(inspect.getsourcefile(f))


def parse_date(string_date):
    """
    >>> now = datetime.now()
    >>> assert now == parse_date(str(now))
    """
    return datetime.strptime(string_date, "%Y-%m-%d %H:%M:%S.%f")


def make_hash(o):
    """
    Returns a hash number for an object, which can also be a dict or a list

    >>> make_hash(range(10))
    -6299899980521991026
    >>> make_hash(list(range(10)))
    -4181190870548101704
    >>> a = make_hash({'a': 1, 'b': 2, 'c': 3})
    >>> b = make_hash({'c': 3, 'a': 1, 'b': 2})
    >>> assert a == b
    """
    def freeze(o):
        if isinstance(o, dict):
            return frozenset({k: freeze(v) for k, v in o.items()}.items())
        if isinstance(o, list):
            return tuple([freeze(v) for v in o])
        return o
    return hash(freeze(o))


def merge(d1, d2):
    """
    Merges two dictionaries, nested values are overwitten by d1

    >>> d = merge({'a': 1}, {'b': 2})
    >>> assert d == {'a': 1, 'b': 2}
    >>> d = merge({'a': {'b': 2}}, {'b': 2, 'a': {'c': 3}})
    >>> assert d == {'a': {'c': 3}, 'b': 2}
    """
    return dict(d1, **d2)


def update_in(d, path, f, *args):
    """
    Parameters:
    -----------
    d: dict
    path: list of key or function
    f: update function (takes nested element or None if element isn't found)

    Applies func `f` on dict `d` on element nested specified by list `path`.
    Items in path are dictionary keys or functions. If a key doesn't match any
    element, an empty dictionary is created at that point. If a nested element
    is a list, the corresponding path item should be a pred func that takes a
    list and returns True at a desired item. If no element in list is matched,
    nothing else happens. If a pred match multiple elements, last one is used.

    # Normal nested update
    >>> d = {'c': {'d': 2}}
    >>> update_in(d, ['c', 'd'], lambda x: x + 3)
    >>> assert d == {'c': {'d': 5}}

    # Update on missing element
    >>> d = {'a': {'b': 1}}
    >>> update_in(d, ['c', 'd'], lambda x: x or [] + [3])
    >>> assert d == {'a': {'b': 1}, 'c': {'d': [3]}}

    # Update on dictionary inside nested list
    >>> d = {'a': [{'b': {'c': 1}}, {'d': 2}]}
    >>> update_in(d, ['a', lambda x: 'b' in x, 'b', 'c'], lambda x: x + 3)
    >>> assert d == {'a': [{'b': {'c': 4}}, {'d': 2}]}

    # Non update on failing pred
    >>> d = {'a': [{'b': {'c': 1}}, {'d': 2}]}
    >>> update_in(d, ['a', lambda x: 'e' in x, 'b', 'c'], lambda x: x + 3)
    >>> assert d == {'a': [{'b': {'c': 1}}, {'d': 2}]}

    # Update on first matching element of list
    >>> d = {'a': [{'b': {'c': 1}}, {'d': 2}, {'b': {'c': 2}}]}
    >>> update_in(d, ['a', lambda x: 'b' in x, 'b', 'c'], lambda x: x + 3)
    >>> assert d == {'a': [{'b': {'c': 1}}, {'d': 2}, {'b': {'c': 5}}]}
    """
    if len(path) == 1:
        if callable(path[0]):
            assert isinstance(d, list), "Found pred but target is %s" % type(d)
            for idx, i in list(enumerate(d))[::-1]:  # reverse list
                if path[0](i):
                    d[idx] = f(i, *args)
                    return
        else:
            d[path[0]] = f(d.get(path[0]), *args)
    else:
        if callable(path[0]):
            assert isinstance(d, list), "Found pred but target is %s" % type(d)
            for i in d[::-1]:
                if path[0](i):
                    update_in(i, path[1:], f, *args)
                    break       # avoid mutating multiple instances
        else:
            if path[0] not in d:
                d[path[0]] = {}
            update_in(d[path[0]], path[1:], f, *args)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import argparse
import numpy as np
from inspect import getsourcefile
from time import time

from keras.models import Model
from keras.layers import Input, Dense
from keras.layers.recurrent import GRU

from casket.experiment import Experiment


def load_data(data, steps=4):
    X, Y = [], []
    for i in range(0, data.shape[0]-steps):
        X.append(data[i: i+steps, :])
        Y.append(data[i+steps, :])
    return np.array(X), np.array(Y)


def train_test_split(data, test_size=0.15):
    X, Y = load_data(data)
    ntrn = round(X.shape[0] * (1 - test_size))
    X_train, Y_train = X[0:ntrn], Y[0:ntrn]
    X_test, Y_test = X[ntrn:], Y[ntrn:]
    return (X_train, Y_train), (X_test, Y_test)


def create_data(item_length, iters):
    data = np.arange(item_length).reshape((item_length, 1))
    for i in range(iters):
        data = np.append(data, data, axis=0)
    return data

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('name')
    parser.add_argument('-l', '--item_length', default=4, type=int)
    parser.add_argument('-e', '--nb_epochs', default=10, type=int)

    args = vars(parser.parse_args())

    params = {                  # experiment_params
        "item_length": args['item_length'],
        "iters": 10,
        "batch_size": 10,
        "nb_epoch": args['nb_epochs'],
        "validation_split": 0.2
    }

    print("Loading data.")
    data = create_data(params["item_length"], params["iters"])
    (X_train, y_train), (X_test, y_test) = train_test_split(data)

    print("Compiling model.")
    in_layer = Input(shape=(4, 1,))
    gru = GRU(10, dropout_W=0.2, dropout_U=0.2)(in_layer)
    dense = Dense(1, activation='linear')(gru)
    model = Model(input=in_layer, output=dense)
    model.compile(loss="mean_squared_error", optimizer="rmsprop", metrics=['accuracy'])

    exp_id = getsourcefile(lambda: 0)
    model_db = Experiment.use("test.json", exp_id=exp_id) \
                         .model("test-model", model.get_config())
    start = time()

    print("Starting learning.")
    with model_db.session(params, ensure_unique=False) as session:
        for e in range(params["nb_epoch"]):
            losses, batch = [], 0
            while batch < X_train.shape[0]:
                batch_to = max(batch + params["batch_size"], X_train.shape[0])
                X_batch, y_batch = X_train[batch: batch_to], y_train[batch: batch_to]
                batch = batch_to
                loss = model.train_on_batch(X_train, y_train)
                losses.append(loss)
                print(np.mean(losses))
            session.add_epoch(e, {'loss': str(np.mean(losses))})
        _, acc = model.test_on_batch(X_test, y_test)
        session.add_result({'test_acc': str(acc)})
        session.add_meta({'run_time': time() - start})




from __future__ import absolute_import
from .experiment import Experiment

try:
    from .callback import DBCallback
    raise ImportError
except ImportError:
    from warnings import warn
    warn("Keras doesn't seem to be installed in your OS", ImportWarning)

from . import nlp_utils



# coding: utf-8

import pytest

from sftp_storage import SFTPStorage
from tinydb import TinyDB, where


element = {'none': [None, None], 'int': 42, 'float': 3.1415899999999999,
           'list': ['LITE', 'RES_ACID', 'SUS_DEXT'],
           'dict': {'hp': 13, 'sp': 5},
           'bool': [True, False, True, False]}

path = 'manjavacas@serv1:~/test.json'


def test_json():
    # Write contents
    storage = SFTPStorage(path, policy='autoadd')
    storage.write(element)

    # Verify contents
    assert element == storage.read()


def test_json_readwrite():
    """
    Regression test for issue #1
    """
    # Create TinyDB instance
    db = TinyDB(path, policy='autoadd', storage=SFTPStorage)

    item = {'name': 'A very long entry'}
    item2 = {'name': 'A short one'}

    get = lambda s: db.get(where('name') == s)

    db.insert(item)
    assert get('A very long entry') == item

    db.remove(where('name') == 'A very long entry')
    assert get('A very long entry') is None

    db.insert(item2)
    assert get('A short one') == item2

    db.remove(where('name') == 'A short one')
    assert get('A short one') is None


def test_json_invalid_directory():
    with pytest.raises(ValueError):
        with TinyDB('/this/is/an/invalid/path/db.json',
                    policy='autoadd', storage=SFTPStorage):
            pass




#!/usr/bin/env python

import argparse


def remove_where(field, match_fn):
    def transform(element):
        pass


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Remove empty models from a db')
    parser.add_argument('db-path')
    args = parser.parse_args()





from tinydb import TinyDB
from itertools import chain


class DB:
    def __init__(self, path):
        try:
            from sftp_storage import SFTPStorage, WrongPathException
            try:
                self.db = TinyDB(path, policy='autoadd', storage=SFTPStorage)
            except WrongPathException:
                self.db = TinyDB(path)
        except ImportError:
            from warnings import warn
            warn("""`paramiko` doesn't seem to be installed in your OS.
            Remote db access is disabled""", ImportWarning)
            self.db = TinyDB(path)

    def get_experiments(self):
        return self.db.search()

    def get_experiment(self, experiment_id):
        return self.db.get({"id": experiment_id})

    def get_model(self, experiment_id, model_id):
        models = self.get_experiment(experiment_id)["models"]
        for m in models:
            if m["modelId"] == model_id:
                return m

    def get_tags(self):
        return chain(*[exp['tags'] for exp in self.db.search()])

    def get_timestamps(self):
        return [model["meta"]["timestamp"]
                for exp in self.get_experiments()
                for model in exp.get("models", [])]

    def get_last_timestamp(self):
        return max(self.get_timestamps)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import argparse
import numpy as np

from keras.models import Sequential
from keras.layers.core import Activation, Dropout, Dense
from keras.layers.recurrent import GRU

from casket.callback import DBCallback


def load_data(data, steps=4):
    X, Y = [], []
    for i in range(0, data.shape[0]-steps):
        X.append(data[i: i+steps, :])
        Y.append(data[i+steps, :])
    return np.array(X), np.array(Y)


def train_test_split(data, test_size=0.15):
    X, Y = load_data(data)
    ntrn = round(X.shape[0] * (1 - test_size))
    X_train, Y_train = X[0:ntrn], Y[0:ntrn]
    X_test, Y_test = X[ntrn:], Y[ntrn:]
    return (X_train, Y_train), (X_test, Y_test)


def create_data(item_length, iters):
    data = np.arange(item_length).reshape((item_length, 1))
    for i in range(iters):
        data = np.append(data, data, axis=0)
    return data

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('name')
    parser.add_argument('-p1', '--param_1', default=4, type=int)
    parser.add_argument('-e', '--nb_epochs', default=10, type=int)
    parser.add_argument('-t', '--tags', type=str, nargs="+", default=["NN"])

    args = vars(parser.parse_args())

    params = {                  # experiment_params
        "item_length": args['param_1'],
        "iters": 10,
        "batch_size": 10,
        "nb_epoch": args['nb_epochs'],
        "validation_split": 0.2
    }

    db_callback = DBCallback(args['name'],
                             corpus="generated",
                             tags=args["tags"],
                             params=params)

    print("Loading data.")
    data = create_data(params["item_length"], params["iters"])
    (X_train, y_train), (X_test, y_test) = train_test_split(data)

    print("Compiling model.")
    model = Sequential()
    model.add(GRU(10,
                  input_dim=1,
                  return_sequences=False))
    model.add(Dropout(0.2))
    model.add(Dense(1))
    model.add(Activation("linear"))
    model.compile(loss="mean_squared_error", optimizer="rmsprop")

    print("Starting learning.")
    model.fit(X_train, y_train,
              batch_size=params["batch_size"],
              nb_epoch=params["nb_epoch"],
              validation_split=params["validation_split"],
              callbacks=[db_callback])



# coding: utf-8

import contextlib
import logging
from datetime import datetime
from uuid import uuid4
from platform import platform
from getpass import getuser

from tinydb import TinyDB, where

from . import utils
from .git import GitInfo


logger = logging.getLogger(__name__)


def log(msg, level=logging.NOTSET):
    logger.log(level, msg)


class ExistingModelParamsException(Exception):
    pass


"""
TinyDB extra operations
"""


def append(field, item):
    """
    Appends item to a list at value `field`
    """
    def transform(element):
        if field not in element:
            element[field] = [item]
        else:
            element[field].append(item)
    return transform


def append_in(path, item):
    """
    Appends item to a list nested in the matching db entry specified by `path`
    """
    def transform(element):
        utils.update_in(element, path, lambda l: (l or []) + [item])
    return transform


def assign_in(path, item):
    """
    Sets item to a dict nested in the matching db entry specified by `path`
    """
    def transform(element):
        utils.update_in(element, path, lambda d: utils.merge(d or {}, item))
    return transform


def extend(field, item):
    """
    Appends item to a set specified by `path` in the matching db entry
    """
    def transform(element):
        if field not in element:
            element[field] = [item]
        if isinstance(element[field], list):
            if item not in set(element[field]):
                element[field].append(item)
    return transform


def remove(field, item):
    """
    Removes item from list
    """
    def transform(element):
        if field not in element:
            return
        if isinstance(element[field], list):
            element[field] = [x for x in element[field] if x != item]
    return transform


"""
Factory update_in preds for models
"""


def model_pred(model_id):
    def f(model):
        return model["modelId"] == model_id
    return f


def params_pred(params):
    def f(result):
        return result["params"] == params
    return f


class Experiment:
    """
    A class to encapsulate information about a experiment and store
    together different experiment runs.
    The recommended way to identify an experiment is overwriting
    Experiment.get_id in a child class in a way that is dependent
    from the source file. Example:

    class MyExperiment(Experiment):
        def get_id(self):
            return inspect.getsourcefile(self)  # return __file__

    Alternatively, one can pass an id as a constructor parameter. If
    no id is passed, the default behaviour is to generate a random id
    (meaning a new experiment is created for each run).

    Experiments are instantiated with the classmethod Experiment.new,
    which additionally stores the experiment in the database with useful
    metadata or Experiment.use, which additionally makes sure that the
    experiment is stored only the first time.
    Actual experiments are run on models. To add a model to the current
    experiment, instantiate the inner class Model with Experiment.model.

    model = Experiment.use(path).model("modelId", {"type": "SVM"})

    Model instantiates and encapsulates model information, providing
    database persistency. A model_id is required to identify the model.
    It also provides convenience methods to store experiment results for
    both single-result and epoch-based training experiments.
    See Model.session

    Parameters:
    -----------
    path: str, path to the database file backend. A path in a remote
        machine can be specified with syntax:
        username@host:/path/to/remote/file.
    """
    def __init__(self, path, exp_id=None, verbose=False):
        assert path, "Path cannot be the empty string"
        self.level = logging.WARN if verbose else logging.NOTSET
        try:
            from sftp_storage import SFTPStorage, WrongPathException
            try:
                self.db = TinyDB(path, policy='autoadd', storage=SFTPStorage)
                log("Using remote db file [%s]" % path, level=self.level)
            except WrongPathException:
                self.db = TinyDB(path)
        except:
            self.db = TinyDB(path)
            log("Using local file [%s]" % path, level=self.level)

        self.git = GitInfo(self.getsourcefile())
        self.id = exp_id if exp_id else self.get_id()

    def get_id(self):
        return uuid4().hex

    def getsourcefile(self):
        return utils.getsourcefile(lambda: None)

    def exists(self):
        return self.db.get(where("id") == self.id)

    def add_tag(self, tag):
        self.db.update(extend("tags", tag), where("id") == self.id)

    def remove_tag(self, tag):
        return self.db.update(remove("tags", tag), where("id") == self.id)

    def get_models(self):
        experiment = self.db.get(where("id") == self.id)
        return experiment.get("models") if experiment else {}

    @classmethod
    def new(cls, path, exp_id=None, tags=(), **params):
        """
        Stores a new Experiment in the database. Throws an exception if
        experiment already exists.
        """
        exp = cls(path, exp_id=exp_id)
        if exp.exists():
            raise ValueError("Experiment %s already exists" % str(exp.id))
        now, exp_id = str(datetime.now()), exp_id or exp.id
        base = {"id": exp_id, "tags": tags, "models": [], "created": now}
        exp.db.insert(utils.merge(base, params))
        return exp

    @classmethod
    def use(cls, path, exp_id=None, tags=(), **params):
        """
        Only stores a new Experiment if none can be find with given parameters,
        otherwise instantiate the existing one with data from database.
        """
        exp = cls(path, exp_id=exp_id)
        if exp.exists():
            return exp
        else:
            log("Creating new Experiment %s" % str(exp.id))
            return cls.new(path, exp_id=exp_id, tags=tags, **params)

    def model_exists(self, model_id):
        """
        Returns dict or None
        """
        return self.db.get(
            (where("id") == self.id) &
            where("models").any(where("modelId") == model_id))

    def model(self, model_id, model_config={}):
        return self.Model(self, model_id, {"config": model_config})

    class Model:
        def __init__(self, experiment, model_id, model_config):
            self._session_params = None
            self.e = experiment
            self.model_id = model_id
            self.which_model = model_pred(self.model_id)
            self.cond = ((where("id") == experiment.id) &
                         where("models").any(where("modelId") == model_id))
            if not self.exists():
                self._add_default_model(**model_config)

        def _add_default_model(self, **kwargs):
            model = utils.merge({"modelId": self.model_id}, kwargs)
            self.e.db.update(append("models", model), where("id") == self.e.id)

        def _result_meta(self):
            return {"commit": self.e.git.get_commit() or "not-git-tracked",
                    "branch": self.e.git.get_branch() or "not-git-tracked",
                    "user": getuser(),
                    "platform": platform(),
                    "timestamp": str(datetime.now())}

        def _check_params(self, params):
            models = self.e.get_models()
            if not models:
                return
            model = next(m for m in models if m["modelId"] == self.model_id)
            for result in model.get("sessions", []):
                if result["params"] == params:
                    raise ExistingModelParamsException()

        def _add_result(self, result, params):
            meta = self._result_meta()
            result = {"params": params, "meta": meta, "result": result}
            path = ["models", self.which_model, "sessions"]
            self.e.db.update(append_in(path, result), self.cond)

        def _add_session_result(self, result):
            """
            Adds (partial) result to session currently running. Session is
            identifed based on session `params`. In case a model is run with
            the same params in a second session, results are added to the
            chronologically last session (which means that we relay on the fact
            that `update_in` checks lists in reverse, see `update_in`)
            """
            which_result = params_pred(self._session_params)
            path = ["models", self.which_model, "sessions",
                    which_result, "result", "epochs"]
            self.e.db.update(append_in(path, result), self.cond)

        def _start_session(self, params):
            self._session_params = params
            path = ["models", self.which_model, "sessions"]
            result = {"params": params, "meta": self._result_meta()}
            self.e.db.update(append_in(path, result), self.cond)

        def _end_session(self):
            self._session_params = None

        def exists(self):
            return self.e.model_exists(self.model_id)

        @contextlib.contextmanager
        def session(self, params, ensure_unique=True):  # to try: store on exit
            """
            Context manager for cases in which we want to add several results
            to the same experiment run. Current session is identified based on
            `params` (see _add_session_result).

            Example:
            model_db = Experiment.use("test.json").model("id")
            with model_db.session({"param-1": 10, "param-2": 100}) as session:
                from time import time
                start_time = time()
                svm.fit(X_train, y_train)
                end_time = time()
                session.add_meta({"duration": end_time - start_time})
                y_pred = svm.predict(X_test)
                session.add_result({"accuracy": accuracy(y_pred, y_true)})

            Parameters:
            -----------
            params: dict, parameters passed in to the model instance
            ensure_unique: bool, throw an exception in case model has already
                been run with the same parameters
            """
            assert isinstance(params, dict), \
                "Params expected dict but got %s" % str(type(params))
            if ensure_unique:
                self._check_params(params)
            self._start_session(params)
            yield self
            self._end_session()

        def add_meta(self, d):
            """
            Adds session meta info

            Parameters:
            -----------
            d: dict, Specifies multiple key-val additional info for the session
            """
            if not self._session_params:
                raise ValueError("add_meta requires session context manager")
            if not isinstance(d, dict):
                raise ValueError("add_meta input must be dict")
            if not self.exists():
                self._add_default_model()
            which_result = params_pred(self._session_params)
            path = ["models", self.which_model,
                    "sessions", which_result, "meta"]
            self.e.db.update(assign_in(path, d), self.cond)

        def add_result(self, result, params=None):
            if not params and not self._session_params:
                raise ValueError("Experiment params missing")
            if not self._session_params:
                self._add_result(result, params)
            else:
                self._add_session_result(result)

        def add_epoch(self, epoch_num, result, timestamp=True):
            if not self._session_params:
                raise ValueError("add_epoch requires session context manager")
            result.update({"epoch_num": epoch_num})
            if timestamp:
                result.update({"timestamp": str(datetime.now())})
            self._add_session_result(result)


if __name__ == '__main__':
    import doctest
    doctest.testmod()




try:
    from casket.git import GitInfo
    version = GitInfo('.').get_tag()
except:
    raise RuntimeError("Couldn't find project tag/version")

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

url = 'https://www.github.com/emanjavacas/casket/tarball/%s' % version

setup(
    name='casket',
    version=version,
    install_requires=[
        'tinydb>=3.2.1'
    ],
    packages=['casket', 'casket.nlp_utils'],
    url='https://www.github.com/emanjavacas/casket',
    download_url=url,
    description='Persistent storage for ML experiments',
    author='Enrique Manjavacas',
    author_email='enrique.manjavacas@gmail.com',
    keywords=['experiments', 'Machine Learning'],
    classifiers=[]
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import json

from .utils import silence

with silence():
    from keras.callbacks import Callback


class DBCallback(Callback):
    """
    Stores and updates a model experiment. A model_id should be used
    for each new model architecture, this enforces that experiments
    based on the same architecture are stored together.
    Parameters
    ----------
    model: experiment.Experiment.Model

    params: dict
        A dictionary of parameters used in the current experiment run.

    freq: int, optional, default 1
        Number of epochs before commiting results to database.
        Can be use to skip over epochs.

    root: str, optional, default 'http://localhost:5000'
        URL to publish results to. Set to None if local server isn't running
    """
    def __init__(self, model, params, freq=1, root='http://localhost:5000'):
        self.model = model
        self.params = params
        self.freq = freq
        self.root = root
        super(Callback, self).__init__()

    def reach_server(self, data, endpoint):
        "tries to reach server at the given `endpoint` with the given `data`"
        if not self.root:
            return
        import requests
        try:
            requests.post(self.root + endpoint, {'data': json.dumps(data)})
        except TypeError:
            print("JSON error; data: " + str(data))
        except:
            print("Could not reach server at " + self.root)

    def on_epoch_begin(self, epoch, logs={}):
        if (epoch % self.freq) == 0:
            self.seen = 0
            self.totals = {}

    def on_epoch_end(self, epoch, logs={}):
        if (epoch % self.freq) == 0:
            epoch_data = {}
            for k, v in self.totals.items():
                epoch_data[k] = v / self.seen
            for k, v in logs.items():  # val_...
                epoch_data[k] = v
            # send to db
            self.model.add_epoch(epoch, epoch_data)

            # send to localhost
            self.reach_server(
                {'epochData': epoch_data, 'modelId': self.model.model_id},
                '/publish/epoch/end/')

    def on_batch_begin(self, batch, logs={}):
        pass

    def on_batch_end(self, batch, logs={}):
        batch_size = logs.get('size', 0)
        self.seen += batch_size
        for k, v in logs.items():  # batch, size
            if k in self.totals:
                self.totals[k] += v * batch_size
            else:
                self.totals[k] = v * batch_size

    def on_train_begin(self, logs={}):
        self.model._start_session(self.params)
        self.reach_server({'action': 'start', 'modelId': self.model_id},
                          '/publish/train/')

    def on_train_end(self, logs={}):
        self.model._end_session()
        self.reach_server({'action': 'end', 'modelId': self.model.model_id},
                          '/publish/train/')




import os
import types


def lines_from_file(fname):
    with open(fname, 'r') as f:
        for line in f:
            yield line


def lines_from_root(root):
    if isinstance(root, types.GeneratorType):
        for line in root:
            yield line
    elif isinstance(root, str):
        if os.path.isdir(root):
            for f in os.listdir(root):
                for l in lines_from_file(os.path.join(root, f)):
                    yield l
    elif os.path.isfile(root):
        for line in lines_from_file(root):
            yield line
    else:
        raise ValueError("Unknown root type [%s]" % type(root))


def pad(items, maxlen, paditem=0, paddir='left'):
    """
    Parameters:
    -----------
    items: iterable, an iterable of objects to index
    maxlen: int, length to which input will be padded
    paditem: any, element to use for padding
    paddir: ('left', 'right'), where to add the padding
    """
    n_items = len(items)
    if n_items == maxlen:
        return items
    if paddir == 'left':
        return (maxlen - n_items) * [paditem] + items
    if paddir == 'right':
        return items + (maxlen - n_items) * [paditem]
    else:
        raise ValueError("Unknown pad direction [%s]" % str(paddir))


class Corpus(object):
    def __init__(self, root, context=10, side='both'):
        """
        Parameters:
        -----------
        root: str/generator, source of lines. If str, it is coerced to file/dir
        context: int, context characters around target char
        side: str, one of 'left', 'right', 'both', context origin
        """
        self.root = root
        self.context = context
        if side not in {'left', 'right', 'both'}:
            raise ValueError('Invalid side value [%s]' % side)
        self.side = side

    def _pad_encode(self, line, indexer, **kwargs):
        """
        Parameters:
        -----------
        line: generator/list, a seq of units to be padded/encoded
        indexer: Indexer, a fitted indexer
        kwargs: optional arguments for Indexer.encode

        Returns:
        --------
        generator over (context, target)
        """
        maxlen = len(line)
        encoded_line = [indexer.encode(c, **kwargs) for c in line]
        for idx, c in enumerate(encoded_line):
            minidx = max(0, idx - self.context)
            maxidx = min(maxlen, idx + self.context + 1)
            if self.side in {'left', 'both'}:
                left = pad(encoded_line[minidx: idx], self.context,
                           paditem=indexer.pad_code, paddir='left')
            if self.side in {'right', 'both'}:
                right = pad(encoded_line[idx + 1: maxidx], self.context,
                            paditem=indexer.pad_code, paddir='right')
            if self.side == 'left':
                yield left, c
            elif self.side == 'right':
                yield right, c
            else:
                yield left + right, c

    def chars(self):
        for line in lines_from_root(self.root):
            for c in line:
                yield c

    def words(self, tokenizer=None):
        for line in lines_from_root(self.root):
            if tokenizer is not None:
                for word in tokenizer(line):
                    yield word
            else:
                for word in line.split():
                    yield word

    def generate(self, mode='chars', tokenizer=None, indexer=None, **kwargs):
        """
        Returns:
        --------
        generator (list:int, int) over instances
        """
        for line in lines_from_root(self.root):
            if mode == 'chars':
                for c, t in self._pad_encode(line, indexer, **kwargs):
                    yield c, t
            elif mode == 'words':
                words = tokenizer(line) if tokenizer else line.split()
                for word, t in self._pad_encode(words, indexer, **kwargs):
                    yield word, t
            else:
                raise ValueError("Unknown mode [%s]" % str(mode))

    def generate_batches(self, batch_size=128, **kwargs):
        """
        Parameters:
        -----------
        batch_size: int
        kwargs: optional arguments for Corpus.generate and Indexer.encode

        Returns:
        --------
        generator (list:list:int, list:int) over batches of instances
        """
        contexts, targets, n = [], [], 0
        generator = self.generate(**kwargs)
        while True:
            try:
                context, target = next(generator)
                if n % batch_size == 0 and contexts:
                    yield contexts, targets
                    contexts, targets = [], []
                else:
                    contexts.append(context)
                    targets.append(target)
                    n += 1
            except StopIteration:
                break



# coding: utf-8

import os
import json
from getpass import getpass

from paramiko import SSHClient, AutoAddPolicy
from tinydb import Storage


class WrongPathException(Exception):
    pass


def parse_url(url):
    """
    Extracts username, host and filename from a scp like url.
    >>> parse_url('johndoe@localhost:~/filename.txt')
    ('johndoe', 'localhost', '~/filename.txt')
    """
    if '@' not in url:
        raise WrongPathException("Bad url format: missing host")
    username, url = url.split("@")
    if not username:
        raise WrongPathException("Couldn't parse url: missing username")
    host, filename = url.split(':')
    return username, host, filename


def find_home(ssh):
    stdin, stdout, stderr = ssh.exec_command("echo $HOME")
    return stdout.readlines()[0].strip()


class SFTPStorage(Storage):
    def __init__(self, path, password=None, policy='default', **kwargs):
        self.username, self.host, self.path = parse_url(path)
        self.kwargs = kwargs
        ssh = SSHClient()
        ssh.load_system_host_keys()
        if policy == 'autoadd':
            ssh.set_missing_host_key_policy(AutoAddPolicy())
        password = password or getpass(
            'Password for %s@%s: ' % (self.username, self.host))
        ssh.connect(self.host, username=self.username, password=password)
        self.ssh = ssh
        self.sftp = ssh.open_sftp()
        if self.path.startswith('~'):
            self.path = os.path.join(find_home(self.ssh), self.path[2:])
        self.sftp.open(self.path, mode='a').close()
        self._handle = self.sftp.open(self.path, mode='r+')

    def read(self):
        self._handle.seek(0, 2)
        size = self._handle.tell()
        if not size:
            return None
        else:
            self._handle.seek(0)
            return json.loads(self._handle.read().decode('utf-8'))

    def write(self, data):
        self._handle.seek(0)
        serialized = json.dumps(data, **self.kwargs)
        self._handle.write(serialized)
        self._handle.flush()
        self._handle.truncate(self._handle.tell())

    def close(self):
        self._handle.close()
        self.ssh.close()
        self.sftp.close()

